/* GStreamer
 * OSX video sink
 * Copyright (C) <2004> Zaheer Abbas Merali <zaheerabbas at merali dot org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

/* Object header */
#include "osxvideosink.h"

#ifdef HAVE_OSX
#import "cocoawindow.h"
#endif

/* Debugging category */
#include <gst/gstinfo.h>
#include <gst/gstelement.h>
GST_DEBUG_CATEGORY_STATIC (gst_debug_osxvideosink);
#define GST_CAT_DEFAULT gst_debug_osxvideosink

static void gst_osxvideosink_dispose (GObject * object);
static gboolean gst_osxvideosink_set_caps (GstBaseSink *sink, GstCaps *caps);
static gboolean gst_osxvideosink_start (GstBaseSink *sink);
static gboolean gst_osxvideosink_stop (GstBaseSink *sink);
static GstFlowReturn gst_osxvideosink_render (GstBaseSink *sink,
    GstBuffer * buffer);
static void gst_osxvideosink_get_times (GstBaseSink *bsink, GstBuffer *buffer,
    GstClockTime *start, GstClockTime *end);
static void gst_osxvideosink_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_osxvideosink_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

/* ElementFactory information */
static GstElementDetails gst_osxvideosink_details =
GST_ELEMENT_DETAILS ("Video sink",
    "Sink/Video",
    "OSX native videosink",
    "Zaheer Abbas Merali <zaheerabbas at merali dot org>");

static GstStaticPadTemplate gst_osxvideosink_sink_template_factory =
GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("video/x-raw-yuv,"
      "framerate=(fraction)[0/1,2147483647/1],"
      "width=(int)[1,2147483647],"
      "height=(int)[1,2147483647],"
      "format=(fourcc)YUY2"
      )
    );


enum
{
  ARG_0,
  ARG_EMBED,
  ARG_FULLSCREEN
      /* FILL ME */
};

enum
{
  SIGNAL_VIEW_CREATED,
  LAST_SIGNAL
};

static guint gst_osxvideosink_signals[LAST_SIGNAL] = { 0 };

GST_BOILERPLATE (GstOSXVideoSink, gst_osxvideosink, GstVideoSink,
    GST_TYPE_VIDEO_SINK);

static void
gst_osxvideosink_base_init (gpointer g_class)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);

  GST_DEBUG_CATEGORY_INIT (gst_debug_osxvideosink, "osxvideosink", 0,
      "osxvideosink element");

  gst_element_class_set_details (element_class, &gst_osxvideosink_details);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_osxvideosink_sink_template_factory));
}

static void
gst_osxvideosink_class_init (GstOSXVideoSinkClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseSinkClass *bs_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;
  bs_class = (GstBaseSinkClass *) klass;

  parent_class = g_type_class_ref (GST_TYPE_VIDEO_SINK);

  gobject_class->dispose = gst_osxvideosink_dispose;
  gobject_class->set_property = gst_osxvideosink_set_property;
  gobject_class->get_property = gst_osxvideosink_get_property;

  g_object_class_install_property (gobject_class, ARG_EMBED,
      g_param_spec_boolean ("embed", "embed", "When enabled, it  "
          "can be embedded", FALSE,
          G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_FULLSCREEN,
      g_param_spec_boolean ("fullscreen", "fullscreen", "When enabled, the view  "
          "is fullscreen", FALSE,
          G_PARAM_READWRITE));

  gst_osxvideosink_signals[SIGNAL_VIEW_CREATED] = g_signal_new("view-created",
    G_TYPE_FROM_CLASS(klass), G_SIGNAL_RUN_CLEANUP, G_STRUCT_OFFSET (GstOSXVideoSinkClass, view_created), NULL, NULL, g_cclosure_marshal_VOID__POINTER, G_TYPE_NONE, 1, G_TYPE_POINTER);

  bs_class->set_caps = gst_osxvideosink_set_caps;
  bs_class->get_times = gst_osxvideosink_get_times;
  bs_class->start = gst_osxvideosink_start;
  bs_class->stop = gst_osxvideosink_stop;
  bs_class->preroll = gst_osxvideosink_render;
  bs_class->render = gst_osxvideosink_render;
}

static void
gst_osxvideosink_init (GstOSXVideoSink * osxvideosink,
    GstOSXVideoSinkClass *klass)
{
  osxvideosink->osxwindow = NULL;
  osxvideosink->osximage = NULL;
  osxvideosink->cur_image = NULL;

  osxvideosink->fps_n = 0;
  osxvideosink->fps_d = 0;

  osxvideosink->pixel_width = osxvideosink->pixel_height = 1;

  osxvideosink->sw_scaling_failed = FALSE;
  osxvideosink->embed = FALSE;
  osxvideosink->fullscreen = FALSE;

}

static void
gst_osxvideosink_dispose (GObject * object)
{
  GstOSXVideoSink *osxvideosink;

  osxvideosink = GST_OSXVIDEOSINK (object);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
gst_osxvideosink_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstOSXVideoSink *osxvideosink;

  g_return_if_fail (GST_IS_OSXVIDEOSINK (object));

  osxvideosink = GST_OSXVIDEOSINK (object);

  switch (prop_id) {
    case ARG_EMBED:
      osxvideosink->embed = g_value_get_boolean(value);
      break;
    case ARG_FULLSCREEN:
      osxvideosink->fullscreen = g_value_get_boolean(value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_osxvideosink_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstOSXVideoSink *osxvideosink;

  g_return_if_fail (GST_IS_OSXVIDEOSINK (object));

  osxvideosink = GST_OSXVIDEOSINK (object);

  switch (prop_id) {
    case ARG_EMBED:
      g_value_set_boolean(value,osxvideosink->embed);
      break;
    case ARG_FULLSCREEN:
      g_value_set_boolean(value,osxvideosink->fullscreen);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}



/* cocoa event loop - needed if not run in own app */
gpointer cocoa_event_loop(gpointer crap)
{
   GST_DEBUG("About to start cocoa event loop");
#ifdef HAVE_OSX
   [NSApp run];
#endif
 
   GST_DEBUG("Cocoa event loop ended");
   return NULL;
}


/* This function handles osx window creation */
static GstOSXWindow *
gst_osxvideosink_osxwindow_new (GstOSXVideoSink * osxvideosink, gint width, gint height)
{
  GstOSXWindow *osxwindow = NULL;
 
  g_return_val_if_fail (GST_IS_OSXVIDEOSINK (osxvideosink), NULL);

  osxwindow = g_new0 (GstOSXWindow, 1);

  osxwindow->width = width;
  osxwindow->height = height;
  osxwindow->internal = TRUE;

  if (osxvideosink->embed == FALSE) {
#ifdef HAVE_OSX
    NSAutoreleasePool *pool;
    NSRect rect;

    rect.origin.x=100.0;
    rect.origin.y=100.0;
    rect.size.width=(float)osxwindow->width;
    rect.size.height=(float)osxwindow->height;

    pool = [[NSAutoreleasePool alloc] init];
    [NSApplication sharedApplication];
    osxwindow->win = [[GstWindow alloc] initWithContentRect:rect styleMask: NSTitledWindowMask|NSClosableWindowMask|NSMiniaturizableWindowMask backing: NSBackingStoreBuffered defer: NO screen: nil];
    [osxwindow->win makeKeyAndOrderFront:NSApp];
    osxwindow->gstview = [osxwindow->win gstView];
    if (osxvideosink->fullscreen) [osxwindow->gstview setFullScreen: YES];
    [pool release];
#endif
    g_thread_create(cocoa_event_loop, osxvideosink, FALSE, NULL);
  } else {
#ifdef HAVE_OSX
    /* Needs to be embedded */
    NSRect rect;
   
    rect.origin.x=0.0;
    rect.origin.y=0.0;
    rect.size.width=(float)osxwindow->width;
    rect.size.height=(float)osxwindow->height;
    osxwindow->gstview = [[GstGLView alloc] initWithFrame:rect];
    /* send signal */
    g_signal_emit (G_OBJECT(osxvideosink),
                   gst_osxvideosink_signals[SIGNAL_VIEW_CREATED], 0,
                   osxwindow->gstview); 
#endif
  }
  return osxwindow;
}

/* This function destroys a GstXWindow */
static void
gst_osxvideosink_osxwindow_destroy (GstOSXVideoSink * osxvideosink,
    GstOSXWindow * osxwindow)
{
  g_return_if_fail (osxwindow != NULL);
  g_return_if_fail (GST_IS_OSXVIDEOSINK (osxvideosink));


  g_free (osxwindow);
}

/* This function resizes a GstXWindow */
static void
gst_osxvideosink_osxwindow_resize (GstOSXVideoSink * osxvideosink, GstOSXWindow * osxwindow,
    guint width, guint height)
{
  g_return_if_fail (osxwindow != NULL);
  g_return_if_fail (GST_IS_OSXVIDEOSINK (osxvideosink));

  //SizeWindow (osxwindow->win, width, height, 1);
  osxwindow->width = width;
  osxwindow->height = height;

  /* Call relevant cocoa function to resize window */
}

static void
gst_osxvideosink_osxwindow_clear (GstOSXVideoSink * osxvideosink, GstOSXWindow * osxwindow)
{

  g_return_if_fail (osxwindow != NULL);
  g_return_if_fail (GST_IS_OSXVIDEOSINK (osxvideosink));

}

#if 0
static void
gst_osxvideosink_osxwindow_update_geometry (GstOSXVideoSink * osxvideosink,
    GstOSXWindow * osxwindow)
{

  g_return_if_fail (osxwindow != NULL);
  g_return_if_fail (GST_IS_OSXVIDEOSINK (osxvideosink));

}

static void
gst_osxvideosink_renegotiate_size (GstOSXVideoSink * osxvideosink)
{
  g_return_if_fail (GST_IS_OSXVIDEOSINK (osxvideosink));

  if (!osxvideosink->osxwindow)
    return;

  gst_osxvideosink_osxwindow_update_geometry (osxvideosink, osxvideosink->osxwindow);

  if (osxvideosink->sw_scaling_failed)
    return;

  if (osxvideosink->osxwindow->width <= 1 || osxvideosink->osxwindow->height <= 1)
    return;

  if (GST_PAD_IS_NEGOTIATING (GST_VIDEOSINK_PAD (osxvideosink)) ||
      !gst_pad_is_negotiated (GST_VIDEOSINK_PAD (osxvideosink)))
    return;


  if (GST_VIDEO_SINK_WIDTH (osxvideosink) != osxvideosink->osxwindow->width ||
      GST_VIDEO_SINK_HEIGHT (osxvideosink) != osxvideosink->osxwindow->height) {
    GstPadLinkReturn r;

    r = gst_pad_try_set_caps (GST_VIDEOSINK_PAD (osxvideosink),
        gst_caps_new_simple ("video/x-raw-yuv",
            "format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'),
            "width", G_TYPE_INT, osxvideosink->osxwindow->width,
            "height", G_TYPE_INT, osxvideosink->osxwindow->height,
            "framerate", G_TYPE_DOUBLE, osxvideosink->framerate, NULL));

    if ((r == GST_PAD_LINK_OK) || (r == GST_PAD_LINK_DONE)) {
      GST_VIDEO_SINK_WIDTH (osxvideosink) = osxvideosink->osxwindow->width;
      GST_VIDEO_SINK_HEIGHT (osxvideosink) = osxvideosink->osxwindow->height;

      if ((osxvideosink->osximage) &&
          ((GST_VIDEO_SINK_WIDTH (osxvideosink) != osxvideosink->osximage->width) ||
              (GST_VIDEO_SINK_HEIGHT (osxvideosink) !=
                  osxvideosink->osximage->height))) {
        gst_osxvideosink_osximage_destroy (osxvideosink, osxvideosink->osximage);

        osxvideosink->osximage = gst_osxvideosink_osximage_new (osxvideosink,
            GST_VIDEO_SINK_WIDTH (osxvideosink),
            GST_VIDEO_SINK_HEIGHT (osxvideosink));
      }
    } else {
      osxvideosink->sw_scaling_failed = TRUE;
    }
  }
}
#endif

/* Element stuff */

static gboolean
gst_osxvideosink_set_caps (GstBaseSink *sink, GstCaps *caps)
{
  GstOSXVideoSink *osxvideosink;
  GstStructure *structure;

  osxvideosink = GST_OSXVIDEOSINK (sink);

  structure = gst_caps_get_structure (caps, 0);

  gst_structure_get_int (structure, "width", &osxvideosink->width);
  gst_structure_get_int (structure, "height", &osxvideosink->height);
  gst_structure_get_fourcc (structure, "format", &osxvideosink->format);
  gst_structure_get_fraction (structure, "framerate",
      &osxvideosink->fps_n, &osxvideosink->fps_d);

  osxvideosink->pixel_width = 1;
  osxvideosink->pixel_height = 1;
  gst_structure_get_fraction (structure, "pixel_ratio",
      &osxvideosink->pixel_width, &osxvideosink->pixel_height);

  return TRUE;
}

static gboolean
gst_osxvideosink_start (GstBaseSink *sink)
{
  GstOSXVideoSink *osxvideosink;

  osxvideosink = GST_OSXVIDEOSINK (sink);

  /* Creating our window and our image */
  if (!osxvideosink->osxwindow) {
    osxvideosink->osxwindow = gst_osxvideosink_osxwindow_new (osxvideosink,
        GST_VIDEO_SINK_WIDTH (osxvideosink), GST_VIDEO_SINK_HEIGHT (osxvideosink));
    gst_osxvideosink_osxwindow_clear (osxvideosink, osxvideosink->osxwindow);
  }
  else {
    if (osxvideosink->osxwindow->internal)
      gst_osxvideosink_osxwindow_resize (osxvideosink, osxvideosink->osxwindow,
          GST_VIDEO_SINK_WIDTH (osxvideosink), GST_VIDEO_SINK_HEIGHT (osxvideosink));
  }

  if (osxvideosink->osxwindow) {
    gst_osxvideosink_osxwindow_clear (osxvideosink, osxvideosink->osxwindow);
  }

  osxvideosink->time = 0;

  return TRUE;
}

static gboolean
gst_osxvideosink_stop (GstBaseSink *sink)
{
  GstOSXVideoSink *osxvideosink;

  /* FIXME this function might not do a thorough cleanup */

  osxvideosink = GST_OSXVIDEOSINK (sink);

  if (osxvideosink->osxwindow) {
    gst_osxvideosink_osxwindow_destroy (osxvideosink, osxvideosink->osxwindow);
    osxvideosink->osxwindow = NULL;
  }

  return TRUE;
}

static GstFlowReturn
gst_osxvideosink_render (GstBaseSink *sink, GstBuffer * buffer)
{
  GstOSXVideoSink *osxvideosink;

  g_return_val_if_fail (sink != NULL, GST_FLOW_ERROR);
  g_return_val_if_fail (buffer != NULL, GST_FLOW_ERROR);
	
  osxvideosink = GST_OSXVIDEOSINK (sink);

#ifdef HAVE_OSX
  char* viewdata = [osxvideosink->osxwindow->gstview getTextureBuffer];
  memcpy(viewdata, GST_BUFFER_DATA (buffer), GST_BUFFER_SIZE(buffer));
  [osxvideosink->osxwindow->gstview displayTexture];
#endif

  return GST_FLOW_OK;
}

static void
gst_osxvideosink_get_times (GstBaseSink *bsink, GstBuffer *buffer,
    GstClockTime *start, GstClockTime *end)
{
  GstOSXVideoSink *osxvideosink;

  osxvideosink = GST_OSXVIDEOSINK (bsink);

  if (GST_BUFFER_TIMESTAMP_IS_VALID (buffer)) {
    *start = GST_BUFFER_TIMESTAMP (buffer);
    if (GST_BUFFER_DURATION_IS_VALID (buffer)) {
      *end = *start + GST_BUFFER_DURATION (buffer);
    } else {
      if (osxvideosink->fps_n > 0) {
        *end = *start +
          gst_util_uint64_scale_int (GST_SECOND, osxvideosink->fps_d, osxvideosink->fps_n);
      }
    }
  }
}



static gboolean
plugin_init (GstPlugin * plugin)
{

  if (!gst_element_register (plugin, "osxvideosink",
          GST_RANK_PRIMARY, GST_TYPE_OSXVIDEOSINK))
    return FALSE;

  return TRUE;
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "osxvideo",
    "OSX native video output plugin",
    plugin_init, VERSION, GST_LICENSE, "GStreamer", "http://gstreamer.freedesktop.org/")
